import { JSXNode, JSXElement, Attributes, Children } from '../jsx';
import { Path } from './path';
import { Patch } from './patch';
/**
 * Diffs two JSX nodes returning a list of patches to be applied by the
 * renderer to the tree trying to mirror our JSX nodes.
 *
 * The list of patches is to be executed in definition order, starting at
 * index 0, all the way down to the final index.
 *
 * To learn more about the different patches that might be returned, see the
 * `Patch` union type.
 */
export declare function diff(a: JSXNode, b: JSXNode, path?: Path): Patch[];
/**
 * Diffs two JSX nodes. If they are both JSX elements, more complex diffing
 * will be done. Otherwise there is just a raw `NodeReplacePatch` performed.
 */
export declare function diffNode(a: JSXNode, b: JSXNode, path: Path): Patch[];
/**
 * Diffs two JSX elements. If the element names are different, a
 * `NodeReplacePatch` is scheduled to be executed. This is because we expect
 * elements of different names to have different behaviors. Therefore we can’t
 * simply swap out the element name, instead we have to replace the full
 * element.
 */
export declare function diffElement(a: JSXElement, b: JSXElement, path: Path): Patch[];
/**
 * Diffs two sets of attributes. This function consists of a fairly simple
 * object comparison algorithm.
 *
 * A few internal attributes (such as `key`) will be ignored.
 */
export declare function diffAttributes(a: Attributes, b: Attributes, path: Path): Patch[];
/**
 * Diffs two lists of JSX nodes which are the children of another JSX node. The
 * algorithm executes the following steps. Relevant patch types are added for
 * reference sake:
 *
 * 1. Construct an index from the old node list where the node’s key is
 *    mapped to the the key’s node.
 * 2. Loop through all of the new nodes and do the following:
 *    1. Get the last node for the same key of this new node.
 *    2. If a last node does not exist add the new node (`ChildrenAddPatch`).
 *    3. If a last node does exist, diff it with the new node.
 * 3. Remove old nodes whose keys do not appear with the new nodes
 *    (`ChildrenRemovePatch`).
 * 4. Order the nodes using the ordered keys from the new nodes
 *    (`ChildrenOrderPatch`).
 */
export declare function diffChildren(a: Children, b: Children, path: Path): Patch[];
